# PyCDDT - Python Custom Debian Distributions Toolkit
# ---------------------
# -*- coding: utf-8 -*-
# ---------------------
# Module:    PyCDDT
# File:      CDDTask.py
# Author:    Sergio Talens-Oliag <sto@debian.org>
# Copyright: (c) 2005, 2006 Sergio Talens-Oliag <sto@debian.org>
# SVN Id:    $Id: CDDTask.py 202 2006-01-28 17:11:39Z sto $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston MA 02110-1301 USA
# ---------------------

"""
Classes to represent Task and Package data from a CDD Description
"""

# -------
# IMPORTS
# -------
import string
import Utils
# L10N Support
from __init__ import _
# Global module variables
from globals import keyDescList, kScope, kType, taskKeywords

# -------
# Classes
# -------
# ---------------
# CDD Task & Pkgd
# ---------------
# Exceptions
class InvalidCDDDictKey(Exception):
  """
  Exception raised when trying to use an invalid keyword on a CDDDict
  """
  def __init__(self, keyword, validKeys):
    Exception.__init__(self)
    self.keyword = keyword
    self.validKeys = validKeys
  def __str__(self):
    return _('Invalid Keyword "%s" on CDDDict\nValid keys are: "%s"') % (self.keyword, self.validKeys)
# Base Class
class CDDDict(dict):
  """
  Class to represent a restricted dictionary, the dictionary limits the valid
  keywords to a list given on the constructor
  """
  # Methods
  def __init__(self, name=None, validKeys=[]):
    """Initialize CDDDict, it needs a name and a list of validKeys"""
    self._name = name
    self._validKeywords = validKeys
  # Restricted dictionary emulation; disallow insertion of invalid keys
  # (except the x- ones)
  def __setitem__(self, key, value):
    """Set item checking if the key is valid"""
    if key not in self._validKeywords:
      if key[0]!='x' or key[1]!='-':
        raise InvalidCDDDictKey(key, self._validKeywords)
    dict.__setitem__(self, key, value)
# CDD Task
class InvalidCDDTaskKey(Exception):
  """
  Exception raised when trying to use an invalid task keyword
  """
  def __init__(self, keyword, validKeys):
    Exception.__init__(self)
    self.keyword = keyword
    self.validKeys = validKeys
  def __str__(self):
    return _('Invalid Task Keyword "%s" on CDDTask\nValid keys are: "%s"') % (self.keyword, self.validKeys)
class CDDTask(CDDDict):
  """
  Class to represent a CDD Task
  """
  # Methods
  def __init__(self, tType, name=None):
    """ 
    Initialize CDDTask objects. The first invocation loads the list of
    validKeys on the class, avoiding the need of duplicating them here
    """
    # Validate the task type
    if tType not in taskKeywords:
      raise InvalidCDDTaskKey(tType, taskKeywords)
    else:
      self._type = tType
      validKeys  = [ tType, ]
      for (k, kd) in keyDescList:
        if kd['scope'].has_key(tType):
          validKeys.append(k)
      # Call parent init
      CDDDict.__init__(self, name, validKeys)
  def taskName(self):
    """Return the task name"""
    return self._name
  def taskType(self):
    """Return the task type"""
    return self._type
  def taskFiles(self, bpath=None):
    fdict={}
    keys=[]
    for k in taskKeywords:
      if self.has_key(k):
        keys.append(k)
    for (k, kd) in keyDescList:
      if k in self.keys() and kd['type'] == kType.fpath:
        if bpath == None:
          fdict[k] = self.__getitem__(k)
        else:
          fdict[k] = map(
              Utils.mkRelPATH,
              [ bpath for v in self.__getitem__(k) ],
              self.__getitem__(k)
          )
    return fdict    
  def taskFieldsString(self, xfields=False, bpath=None):
    keys=[]
    for k in taskKeywords:
      if self.has_key(k):
        keys.append(k)
    s=""
    for (k, kd) in keyDescList:
      if k in self.keys():
        if s != "": s += "\n"
        if kd['type'] == kType.pdep:
          value = string.join(map(Utils.pdepToStr, self.__getitem__(k)), ", ")
        elif kd['type'] == kType.plist:
          value = string.join(self.__getitem__(k), ", ")
        elif kd['type'] == kType.fpath:
          filelist = self.__getitem__(k)
          if bpath != None:
            filelist = map(
              Utils.mkRelPATH,
              [ bpath for v in self.__getitem__(k) ],
              self.__getitem__(k)
            )
          value = string.join(filelist, "\n%s: " % (kd['name']))  
        else:
          value = self.__getitem__(k)
        s += "%s: %s" % (kd['name'], value)
    if xfields:
      for k in self.keys():
        if k[0] == 'x' and k[1] == '-' and k not in keys:
          xk = string.capitalize(k)
          for v in self.__getitem__(k):
            if s != "": s += "\n"
            s += "%s: %s" % (xk, v)
    return s
  def __str__(self):
    return self.taskFieldsString()
# -----------
# CDDTaskSet
# -----------
# Exceptions
class InvalidCDDTask(Exception):
  """
  Exception raised when trying to insert something that is not a CDDTask on a
  CDDTaskSet
  """
  def __init__(self, keyword, value):
    Exception.__init__(self)
    self.keyword = keyword
    self.value   = value
  def __str__(self):
    return _('Invalid value "%s" for task "%s" on CDDTaskSet') % (
      self.keyword, self.value
    )
# CDDTaskSet restricted dictionary
class CDDTaskSet(dict):
  """
  Restricted dictionary to store the set of tasks loaded from a description
  file
  """
  def __init__(self):
    # Depends
    self.lastDepKeys = None
    self.taskDeps    = {}
    self.taskDepsStr = {}
  # Restricted dictionary emulation; insert CDDTask objects only
  def __setitem__(self, key, value):
    """Set item checking if the key is valid"""
    if type(value) == type(CDDTask):
      raise InvalidCDDTask(key,value)
    dict.__setitem__(self, key, value)
  # Private function
  def _genDepsTree(self, task, tDepKeys=None):
    """
    Private function that, given a task name, generates an ordered list of
    tasks that have to be installed to verify the task dependencies
    """
    # Initial test
    if task not in self.keys(): return None
    # Depend keywords
    if tDepKeys == None: tDepKeys = [ 'task-depends', ]
    # Test if the already precomputed trees are valid
    if self.lastDepKeys != tDepKeys:
      self.taskDeps    = {}
      self.taskDepsStr = {}
    # Record the keys we are using to generate the tree
    self.lastDepKeys = tDepKeys
    # Gen deplist for a task
    tasks  = [ task, ]
    tlevel = [ 0, ]
    depath = []
    ttree  = []
    tdtStr = ""
    while (len(tasks)):
      ctask = tasks.pop()
      level = tlevel.pop() 
      # See if the task exists
      if ctask not in self.keys():
        raise Exception(_("Unknown task '%s'") % ctask)
      # Beg: dependency loop detection
      # ------------------------------
      # First, remove old path elements
      if len(depath) > level:
        del depath[level:]
      # Check for loops
      if ctask in depath:
        raise Exception(
          _("Dependency loop found for task '%s': %s")
          % (ctask, string.join(depath, ", "))
        )
      # Add new task
      depath.append(ctask)
      # ------------------------------
      # End: dependency loop detection
      # Add task to the ttree iff it is not on the current level
      if len(ttree) == level:
        ttree.append([ctask])
      elif ctask not in ttree[level]:
        ttree[level].append(ctask)
      # Save Tasks Dep Str
      tdtStr += "# "
      for x in xrange(level-1):
        tdtStr += "  "
      if level > 0:
        tdtStr += "|-"
      tdtStr += "> " + ctask + "\n"
      # Compute this task dependencies
      for td in tDepKeys:
        ct = self.__getitem__(ctask)
        if ct.has_key(td):
          for t in ct[td]:
            tasks.append(t)
            tlevel.append(level+1)
    # Once here we have the full task tree ready, we will install it starting
    # from the deepest level and will remove all repeated tasks (there can be
    # repeated tasks at different levels, once one is installed don't install
    # it again)
    tInstOrder = []
    for l in xrange(len(ttree)-1, -1, -1):
      for t in ttree[l]:
        if t not in tInstOrder:
          tInstOrder.append(t) 
    # Check task-conflicts, just in case
    for ctask in tInstOrder:
      ct = self.__getitem__(ctask)
      if ct.has_key('task-conflicts'):
        for t in ct['task-conflicts']:
          if t in tInstOrder:
            raise Exception(
              _("Error in '%s' deps: task '%s' conflicts with task '%s'") %
              (task, ctask, t))
    return tInstOrder, tdtStr
  # Verify the dependency tree for all tasks
  def verifyDepTrees(self, tDepKeys=None):
    """
    Generate, for each task, an ordered list of tasks that have to be
    installed to verify the task dependencies
    """
    # Init taskDeps & taskDepsStr
    for task in self.keys():
      if task not in self.taskDeps:
        (td, tds) = self._genDepsTree(task, tDepKeys)
        self.taskDeps[task]    = td
        self.taskDepsStr[task] = tds
    return True
  def taskDepTree(self, task, tDepKeys=None):
    """
    Get the ordered list of tasks that have to be installed to verify the
    given task dependencies
    """
    if task not in self.keys():
      return None
    if task not in self.taskDeps:
      (td, tds) = self._genDepsTree(task, tDepKeys)
      self.taskDeps[task]    = td
      self.taskDepsStr[task] = tds
    return self.taskDeps[task]
  def taskDepTreeStr(self, task, tDepKeys=None):
    """
    Get the ordered list of tasks that have to be installed to verify the
    given task dependencies
    """
    if task not in self.keys():
      return None
    if task not in self.taskDepsStr:
      (td, tds) = self._genDepsTree(task, tDepKeys)
      self.taskDeps[task]    = td
      self.taskDepsStr[task] = tds
    return self.taskDepsStr[task]

# ------
# SVN Id: $Id: CDDTask.py 202 2006-01-28 17:11:39Z sto $
