# ---------------------------------------------------
# PyCDDT - Python Custom Debian Distributions Toolkit
# ---------------------------------------------------
# Module:    PyCDDT.Commands
# File:      CDDCommand.py
# Author:    Sergio Talens-Oliag <sto@debian.org>
# Copyright: (c) 2005, 2006 Sergio Talens-Oliag <sto@debian.org>
# SVN Id:    $Id: Commands.py 310 2006-02-04 10:46:37Z sto $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston MA 02110-1301 USA
# ---------------------
# -*- coding: utf-8 -*-
# ---------------------

"""
Custom Debian Distributions Tool Commands

This file includes classes that implement the cddtool commands, each class
includes information about the command and it's parammeters and implements
a method to execute the command.
"""

# -------
# IMPORTS
# -------
import getopt
import os
import string
import sys
# Utilities
import Utils
lpout=Utils.lpout
lperr=Utils.lperr

# L10N Support
from __init__ import _
# CDD Classes
from CDDTask  import CDDTaskSet
from CtrlFile import ControlFile
from globals  import keywords, taskKeywords, metaKeywords, kType

# ----------
# Base Class
# ----------

# Exceptions
class CDDCmndError(Exception):
  """
    Exception raised when there is an error parsing a CDDCommand
  """
  def __init__(self, msg):
    Exception.__init__(self)
    self.msg = msg
  def __str__(self):
    return self.msg

class CDDCommand:
  def __init__(self, name, description, options, parameters):
    self.name = name
    self.desc = description
    self.opts = options
    self.prms = parameters
    self.cddGlobals = None
    self.cddTasks   = None
    self.tasksAvail = None

  def usage(self, msg=None):
    param_lst = string.join([p for p,pd in self.prms], " ")
    param_str = string.join(["  %-29s %s"%(p,pd) for p,pd in self.prms],"\n")
    opts_str  = ""
    for short_opt, long_opt, opt_desc in self.opts:
      if len(short_opt) > 0:
        opts_str += "  -%c, --%-23s %s\n" % (short_opt[0], long_opt, opt_desc)
      else:
        opts_str += "      --%-23s %s\n" % (long_opt, opt_desc)
    str = ""
    if msg:
      str += "%s\n" % msg
    str += _("Command usage:\n  %s [OPTIONS]") % (self.name)
    if len(param_lst) > 0:
      str += " %s\n" % (param_lst)
      str += _("\nWhere:\n%s\n") % (param_str)
    str += _("\nOPTIONS:\n%s") % (opts_str)
    return str
  
  def parse_args(self, args):
    # Parse options
    short_opts = ""
    long_opts  = []
    for so, lo, ignore in self.opts:
      if len(so) > 0:
        short_opts += so
      long_opts.append(lo)
    try:
      return getopt.getopt(args, short_opts, long_opts)
    except getopt.GetoptError, error:
      raise CDDCmndError( _("ERROR reading options: %s") % (error))

  def check_cdd(self, globalOpts):
    if globalOpts.has_key('cddGlobals'):
      self.cddGlobals = globalOpts['cddGlobals']
    if globalOpts.has_key('cddTasks'):
      self.cddTasks   = globalOpts['cddTasks']
      self.tasksAvail = globalOpts['tasksAvail']
      if self.tasksAvail == None or len(self.tasksAvail) == 0:
        raise CDDCmndError( _("I have no tasks!") )
    else:
      raise CDDCmndError( _("I need a description file!") )
    return 0

  def run(self, globalOpts, args):
    lpout("%s: Unimplemented command" % self.name)
    return 1

# --------
# Commands
# --------

class Build(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'build', _("Build metapackages, task lists, installation CD..."),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Get(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'get', _("Download packages needed for the CDD"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Install(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'install', _("Install packages related to a task"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Reconfigure(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'reconfigure', _("Reconfigure packages related to a task"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Remove(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'remove', _("Remove packages installed for a task"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Purge(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'purge', _("Purge packages installed for a task"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

class Tdeps(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'tdeps', _("Print task dependencies"), (
        ('r',  'task-recommends', _("Include recommended tasks on deps")),
        ('s',  'task-suggests',   _("Include suggested tasks on deps")),
        ('v',  'verify',          _("Verify all the task dependencies")),
      ), (
        ('TASK_LIST', _("List of tasks to compute dependencies for") ),
      ),
    )
  def run(self, globalOpts, args):
    # Check tasks
    res = self.check_cdd(globalOpts)
    if res != 0: return res
    # Local variables
    pkgDepends  = [ 'conflicts', 'depends', 'recommends', 'suggests', ]
    tskDepends  = [ 'task-depends', ]
    # Parse options
    (opts, xargs) = self.parse_args(args)
    taskList = xargs
    for o, v in opts:
      if o in ('-r', '--task-recommends'):
        if 'task-recomends' not in tskDepends:
          tskDepends.append('task-recommends')
      elif o in ('-s', '--task-suggests'):
        if 'task-suggests' not in tskDepends:
          tskDepends.append('task-suggests')
      elif o in ('-v', '--verify'):
        if self.cddTasks.verifyDepTrees(tskDepends):
          lpout(_("The task dependencies of all current tasks are right"))
          return 0
        else:
          lpout(_("There are errors on the dependencies of the current tasks"))
          return 1
      else:
        raise CDDCmndError( _("Unimplemented option '%s'") % (o) )
    # Check the arguments
    if len(taskList) == 0:
      raise CDDCmndError (
        _("I need at least one task, use 'tinfo -l' to get a list")
      )
    # Show the requested information 
    for task in taskList:
      if task not in self.tasksAvail:
        lpout(_("# Task '%s' not found on the available task list!") % task)
        taskList.remove(task)
    # Evaluate tasks
    tskInclDeps = []
    for task in taskList:
      # Say which task we are processing
      lpout(_("# Task: '%s'") % (task))
      # Get tasks
      tdt    = self.cddTasks.taskDepTree(task, tskDepends)
      tdtStr = self.cddTasks.taskDepTreeStr(task, tskDepends)
      # Print Tasks
      lpout(_("# Needs '%d' tasks: %s") % (len(tdt), string.join(tdt, ', ')))
      print tdtStr,
      # Generate empty deplist
      dl = {}
      for d in pkgDepends: dl[d]=[]
      # Process tasks expanding all dependencies
      for tname in tdt:
        t = self.cddTasks[tname]
        # Process dependencies
        lpout("# Processing task: '%s'" % (tname))
        for pd in pkgDepends:
          if t!=None and t.has_key(pd):
            tdeps=[]
            for dep in t[pd]:
              ds = Utils.pdepToStr(dep)
              # Include dependency if not already present
              if ds not in dl[pd]:
                tdeps.append(ds)
            dl[pd] += tdeps
            if pd in keywords and keywords[pd].has_key('name'):
              pdname = keywords[pd]['name']
            else:
              pdname = string.capwords(pd)
            lpout("%s: %s" % (pdname, string.join(tdeps, ', ')))
    return 0

class Tinfo(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'tinfo', _("Print information about tasks"), (
        ('a',  'all',      _("Dump information for all tasks")),
        ('g',  'gfields',  _("Print CDD global fields")),
        ('l',  'list',     _("Print the list of available tasks")),
        ('m',  'meta',     _("Print meta-tasks and flavours")), 
        ('t:', 'type=',    _("Print tasks of the given type")),
        ('x',  'x-fields', _("Print X- fields")),
      ), (
        ('TASK_LIST', _("List of tasks to compute dependencies for") ),
      ),
    )
  def run(self, globalOpts, args):
    # Check tasks
    res = self.check_cdd(globalOpts)
    if res != 0: return res
    # Parse options
    (opts, xargs) = self.parse_args(args)
    # Local variables
    taskList = xargs
    gfields  = False
    list     = False
    meta     = False
    tType    = []
    xfields  = False
    for o, v in opts:
      if o in ('-a', '--all'):
        taskList = self.tasksAvail
      elif o in ('-g', '--gfields'):
        gfields = True
      elif o in ('-l', '--list'):
        list = True
        taskList = self.tasksAvail
      elif o in ('-m', '--meta'):
        meta = True
        taskList = self.tasksAvail
      elif o in ('-t', '--type'):
        if v in taskKeywords:
          tType.append(v)
        else:
          raise CDDCmndError ("Unknown task type '%s'\nValid values are: %s"\
            % (v, taskKeywords)) 
      elif o in ('-x', '--x-fields'):
        xfields = True
      else:
        raise CDDCmndError ( _("Unimplemented option '%s'") % (o) )
    # Check the arguments
    if len(taskList) == 0:
      raise CDDCmndError( _("I need at least one task, use 'tinfo -l' to get a list") )
      return 1
    # Print information
    if gfields and self.cddGlobals != None and not meta and not list:
      gflist = self.cddGlobals.keys()
      gflist.sort()
      for k in gflist:
        for v in self.cddGlobals[k]:
          lpout("%s: %s" % (string.capitalize(k), v))
      print ""
    for task in taskList:
      if task not in self.tasksAvail:
        lpout(_("# Task '%s' not found on the available task list!") % task)
        taskList.remove(task)
      elif len(tType)>0 and self.cddTasks[task].taskType() not in tType:
        taskList.remove(task)
    for task in taskList:
      t=self.cddTasks[task]
      if meta:
        for mk in metaKeywords:
          if t.has_key(mk) and t[mk]:
            lpout("%s" % (task))
            break
      elif list:
        lpout("%s" % (task))
      else:
        lpout("%s\n" % (t.taskFieldsString(xfields)))
    return 0

class Tfiles(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'tfiles', _("Print files related to a task (and its dependencies)"), (
        ('b:', 'base-path=',  _("Print file paths relative to the given one")),
        ('e',  'expand-meta', _("List files of metatasks I depend on")),
        ('i:', 'install-path=',  _("Install files under that path")),
        ('l',  'list', _("List fields that can contain files")),
        ('r',  'task-recommends', _("Include recommended tasks on deps")),
        ('s',  'task-suggests', _("Include suggested tasks on deps")),
      ), (
        ('TASK', _("Task to print files for") ),
        ('[FIELDS]', _("List of fields to print") ),
      ),
    )
  def run(self, globalOpts, args):
    # Check tasks
    res = self.check_cdd(globalOpts)
    if res != 0: return res
    # Local variables
    base_path    = None
    install_path = None
    expandMeta   = False
    tskDepends   = [ 'task-depends', ]
    # Parse options
    (opts, xargs) = self.parse_args(args)
    for o, v in opts:
      if o in ('-b', '--base-path'):
        base_path = v
      elif o in ('-i', '--install-path'):
        install_path = v
      elif o in ('-e', '--expand-meta'):
        expandMeta = True
      elif o in ('-l', '--list'):
        tl = []
        for k in keywords:
          if keywords[k].has_key('type') and keywords[k]['type'] == kType.file:
            tl.append(k)
        tl.sort()
        lpout(string.join(tl, '\n'))
        return 0
      elif o in ('-r', '--task-recommends'):
        if 'task-recomends' not in tskDepends:
          tskDepends.append('task-recommends')
      elif o in ('-s', '--task-suggests'):
        if 'task-suggests' not in tskDepends:
          tskDepends.append('task-suggests')
      else:
        raise CDDCmndError ( _("Unimplemented option '%s'") % (o) )
    # Check the arguments
    if len(xargs) > 0:
      task = xargs[0]
    else:
      task = None
    if task == None or task not in self.tasksAvail:
      raise CDDCmndError( _("I need a valid task, use 'tinfo -l' to get a list") )
      return 1
    fileFields = xargs[1:]
    if len(fileFields) > 0:
      for f in fileFields:
        if f not in keywords or 'type' not in keywords[f] \
          or keywords[f]['type'] != kType.file:
            raise CDDCmndError( _("Wrong field '%s'") % f)
    else:
      for k in keywords:
        if keywords[k].has_key('type') and keywords[k]['type'] == kType.file:
          fileFields.append(k)
    # Evaluate tasks
    tskInclDeps = []
    if 'task-conflicts' not in tskDepends:
      tskInclDeps.append('task-conflicts')
    tskInclDeps += tskDepends
    # Get tasks
    tdt = self.cddTasks.taskDepTree(task, tskDepends)
    # Print tasks as dependencies
    if not expandMeta:
      for tname in tdt:
        t = self.cddTasks[tname]
        for td in tskInclDeps:
          if t.has_key(td):
            for stname in t[td]:
              st=self.cddTasks[stname]
              for mk in metaKeywords:
                if st.has_key(mk) and st[mk]:
                  # Remove this task and it's dependencies
                  if stname in tdt:
                    tdt.remove(stname)
                    stdt = self.cddTasks.taskDepTree(stname)
                    for stdname in stdt:
                      if stdname in tdt:
                        tdt.remove(stdname)
                  break
    # Process dependencies
    for tname in tdt:
      lpout(_("# Processing task '%s'") % tname)
      t = self.cddTasks[tname]
      fdict = t.taskFiles(base_path)
      for field in fileFields:
        if field in fdict:
          if install_path:
            destdir="\t%s/tasks/%s/%s/"%(install_path, tname, field)
          else:
            destdir="\t$CDDROOT/tasks/%s/%s/"%(install_path, tname, field)
          for filename in fdict[field]:
            lpout("%s%s" % (filename, destdir))
    return 0

# Binary packages control fields:
# -------------------------------
# Package (mandatory)
# Source
# Version (mandatory)
# Section (recommended)
# Priority (recommended)
# Architecture (mandatory)
# Essential
# Depends et al
# Installed-Size
# Maintainer (mandatory)
# Description (mandatory)
class Tmeta(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'tmeta', _("Show metapackages' substvars"), (
        ('c',  'control', _("Print binary control files instead of substvars")),
        ('e',  'expand-meta', _("Don't depend on metatasks, expand them")),
        ('r',  'task-recommends', _("Include recommended tasks on deps")),
        ('s',  'task-suggests', _("Include suggested tasks on deps")),
        ('I:', 'Installed-Size=',
         _("Installed-Size field, defaults to 'env CDDT_META_INSTALLED_SIZE'")),
        ('M:', 'Maintainer=',
         _("Maintainer field, defaults to 'env CDDT_META_MAINTAINER'")),
        ('S:', 'Source=',
         _("Source field, defaults to 'env CDDT_META_SOURCE'")),
        ('V:', 'Version=',
         _("Version field, defaults to 'env CDDT_META_VERSION'")),
        ('',   'add-conflicts=', _("Add given dependency to the task Conflicts")),
        ('',   'add-depends=', _("Add given dependency to the task Depends")),
        ('',   'add-enhaces=', _("Add given dependency to the task Enhaces")),
        ('',   'add-pre-depends=', _("Add given dependency to the task Pre-Depends")),
        ('',   'add-provides=', _("Add given dependency to the task Provides")),
        ('',   'add-recommends=', _("Add given dependency to the task Recommends")),
        ('',   'add-replaces=', _("Add given dependency to the task Replaces")),
        ('',   'add-suggests=', _("Add given dependency to the task Suggests")),
      ), (
        ('META_PREFIX', _("Metapackages prefix (description package name)") ),
        ('TASK_LIST',   _("List of tasks to compute dependencies for") ),
      ),
    )
  def run(self, globalOpts, args):
    # Check tasks
    res = self.check_cdd(globalOpts)
    if res != 0: return res
    # Local variables
    pkgDepends  = [ 'conflicts', 'depends', 'recommends', 'suggests', ]
    tskDepends  = [ 'task-depends', ]
    mpDepends   = [ 'conflicts', 'depends', 'enhaces', 'pre-depends',
                    'provides', 'recommends', 'replaces', 'suggests', ]
    tskFields   = [ 'architecture', 'section', 'priority', 'description', ]
    control     = False
    expandMeta  = False
    defaultDL   = {}
    for md in mpDepends:
      defaultDL[md] = []
    # Environment variables
    metaInstalledSize = ''
    if os.environ.has_key("CDDT_META_INSTALLED_SIZE"):
      metaInstalledSize = os.environ["CDDT_META_INSTALLED_SIZE"]
    metaMaintainer = ''
    if os.environ.has_key("CDDT_META_MAINTAINER"):
      metaMaintainer = os.environ["CDDT_META_MAINTAINER"]
    metaSource = ''
    if os.environ.has_key("CDDT_META_SOURCE"):
      metaSource = os.environ["CDDT_META_SOURCE"]
    metaVersion = ''
    if os.environ.has_key("CDDT_META_VERSION"):
      metaVersion = os.environ["CDDT_META_VERSION"]
    # Parse options
    (opts, xargs) = self.parse_args(args)
    for o, v in opts:
      if o in ('-c', '--control'):
        control = True
      elif o in ('-e', '--expand-meta'):
        expandMeta = True
      elif o in ('-r', '--task-recommends'):
        if 'task-recomends' not in tskDepends:
          tskDepends.append('task-recommends')
      elif o in ('-s', '--task-suggests'):
        if 'task-suggests' not in tskDepends:
          tskDepends.append('task-suggests')
      elif o in ('-I', '--Installed-Size'):
        metaInstalledSize = v
      elif o in ('-M', '--Maintainer'):
        metaMaintainer = v
      elif o in ('-S', '--Source'):
        metaSource = v
      elif o in ('-V', '--Version'):
        metaVersion = v
      elif o[:len('--add-')]=='--add-':
        defaultDL[ o[len('--add-'):] ].append(v)
      else:
        self.usage( _("Unimplemented option '%s'") % (o) )
        return 1
    # Check the arguments
    if len(xargs) < 2:
      raise CDDCmndError( _("I need the META_PREFIX and at least one task!") )
    metaPrefix = xargs[0]
    metaSuffix = ''
    taskList   = xargs[1:]
    # Show the requested information 
    for task in taskList:
      if task not in self.tasksAvail:
        lperr(_("# Task '%s' not found on the available task list!") % (task))
        taskList.remove(task)
    # Put a - at the end of the prefix if it does not have one
    if metaPrefix and metaPrefix[-1] != '-':
      metaPrefix += '-'
    # Put the version on the metaSuffix, if available
    if metaVersion:
      metaSuffix = " (= %s)" % (metaVersion)
    # Evaluate tasks
    tskInclDeps = []
    if 'task-conflicts' not in tskDepends:
      tskInclDeps.append('task-conflicts')
    tskInclDeps += tskDepends
    for tsk in taskList:
      # Say which task we are processing
      if control:
        lpout(_("Package: %s") % (metaPrefix + tsk))
        if metaSource:
          lpout(_("Source: %s") % (metaSource))
        if metaVersion:
          lpout(_("Version: %s") % (metaVersion))
        if metaInstalledSize:
          lpout(_("Installed-Size: %s") % (metaInstalledSize))
        if metaMaintainer:
          lpout(_("Maintainer: %s") % (metaMaintainer))
      else:
        lpout(_("cddt:Package=%s") % (metaPrefix + tsk))
        lpout(_("cddt:Source=%s") % (metaSource))
        lpout(_("cddt:Version=%s") % (metaVersion))
        lpout(_("cddt:Installed-Size=%s") % (metaInstalledSize))
        lpout(_("cddt:Maintainer=%s") % (metaMaintainer))
      # Get tasks
      tdt    = self.cddTasks.taskDepTree(tsk, tskDepends)
      tdtStr = self.cddTasks.taskDepTreeStr(tsk, tskDepends)
      # Generate empty deplist
      dl={}
      dl.update(defaultDL)
      # Print tasks as dependencies
      if not expandMeta:
        for tname in tdt:
          t = self.cddTasks[tname]
          for td in tskInclDeps:
            if t.has_key(td):
              for stname in t[td]:
                st=self.cddTasks[stname]
                for mk in metaKeywords:
                  if st.has_key(mk) and st[mk]:
                    dl[ td[len("task-"):] ].append(
                      metaPrefix + stname + metaSuffix
                    )
                    # Remove this task and it's dependencies
                    if stname in tdt:
                      tdt.remove(stname)
                      stdt = self.cddTasks.taskDepTree(stname)
                      for stdname in stdt:
                        if stdname in tdt:
                          tdt.remove(stdname)
                    break
      # Process dependencies
      for tname in tdt:
        t = self.cddTasks[tname]
        for pd in pkgDepends:
          if t!=None and t.has_key(pd):
            tdeps=[]
            for dep in t[pd]:
              ds = Utils.pdepToStr(dep)
              # Include dependency if not already present
              if ds not in dl[pd]:
                tdeps.append(ds)
            dl[pd] += tdeps
      # Print substvars
      for tf in tskFields:
        if tf in keywords and keywords[tf].has_key('name'):
          tfname = keywords[tf]['name']
        else:
          tfname = string.capwords(tf)
        if t.has_key(tf):
          if control:
            lpout(tfname + ": " + t[tf])
          else:
            lpout("cddt:" + tfname + "=" \
              + string.replace(t[tf], "\n", "${Newline}"))
        elif not control:
          lpout("cddt:" + tfname + "=")
      for pd in dl.keys():
        if pd in keywords and keywords[pd].has_key('name'):
          pdname = keywords[pd]['name']
        else:
          pdname = string.capwords(pd)
        if len(dl[pd]) > 0:
          if control:
            lpout(pdname + ": " + string.join(dl[pd], ', '))
          else:
            lpout("cddt:" + pdname + "=" \
              + string.replace(string.join(dl[pd], ', '), "\n", "${Newline}"))
        elif not control:
          lpout("cddt:" + pdname + "=")
      print ""
    return 0

class Update(CDDCommand):
  def __init__(self):
    CDDCommand.__init__(self,
      'update', _("Update packages related to a task"),
      (
        ('h', 'help', _("Display this help and exit") ), 
      ),
      (
        ('[TASKLIST]', _("List of tasks to build (ALL if missing)")), 
      ),
    )

# ------------------
# Commands to export
# ------------------
commands = {
  'build':       Build(),
  'get':         Get(),
  'install':     Install(),
  'reconfigure': Reconfigure(),
  'remove':      Remove(),
  'purge':       Purge(),
  'tdeps':       Tdeps(),
  'tinfo':       Tinfo(),
  'tfiles':      Tfiles(),
  'tmeta':       Tmeta(),
  'update':      Update(),
}

# ------
# SVN Id: $Id: Commands.py 310 2006-02-04 10:46:37Z sto $
