# ---------------------------------------------------
# PyCDDT - Python Custom Debian Distributions Toolkit
# ---------------------------------------------------
# Module:    PyCDDT
# File:      CtrlFile.py
# Author:    Sergio Talens-Oliag <sto@debian.org>
# Copyright: (c) 2005, 2006 Sergio Talens-Oliag <sto@debian.org>
# SVN Id:    $Id: CtrlFile.py 235 2006-01-29 22:14:54Z sto $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston MA 02110-1301 USA
# ---------------------
# -*- coding: utf-8 -*-
# ---------------------

"""
Classes to read, parse and represent a Custom Debian Distribution Description
"""

# -------
# IMPORTS
# -------
import apt_pkg
import os
import string
# L10N Support
from __init__ import _
# Axiliary clases
from CDDTask import CDDTask, CDDTaskSet
# Global module variables
from globals import kScope, kDup, kType, eType
from globals import keywords
from globals import inclKeywords, taskKeywords
# Auxiliary functions
from Utils import chkPackageName, chkScriptName

# -------
# Classes
# -------
# Exceptions
class InvalidCDDDictKey(Exception):
  """
  Exception raised when trying to use an invalid keyword on a CDDDict
  """
  def __init__(self, keyword):
    Exception.__init__(self)
    self.keyword = keyword
  def __str__(self):
    return _('Invalid Keyword "%s" on CDDDict') % self.keyword

class ControlFileError(Exception):
  """
  Exception raised when an error is detected reading a Control File.
  Each error is a tuple (eType, msg).
  """
  def __init__(self, errors):
    Exception.__init__(self)
    self.errors = errors
  def __str__(self):
    return string.join( [ m for (e,m) in self.errors ] , '\n')

# Control file class
class ControlFile:
  """
  Class to represent the control file information of a CDD
  """
  # --------------------
  # Public Class Methods
  # --------------------
  def __init__(self, filename, stopOnFstErr=False):
    self.stopOnFirstError = stopOnFstErr
    self.filename         = os.path.abspath(filename)
    self.globals          = None
    self.tasks            = None
    self.paragraphs       = None

  def parseFile(self):
    """
    Parse the file, removes previous versions of the parsed file
    """
    self.globals    = {}
    self.tasks      = CDDTaskSet()
    self.paragraphs = []
    res = self._readFile()
    if res:
      res = self._parseParagraphs()
    del self.paragraphs
    return res

  def getGlobals(self):
    """
    Returns the dictionary of global keys
    """
    return self.globals

  def getTasks(self):
    """
    Returns the list of tasks on the control file.
    """
    return self.tasks

  # -----------------
  # Private functions
  # -----------------
  def _readFile(self):
    """
    Function that reads the given filename and included files and generates a
    list of paragraphs that is stored on the class.

    This function is a wrapper that calls the real ``__readFile`` function and
    is defined only to be able to raise an exception with all the errors found
    when ``stopOnFirstError`` is False (the errors are stored on a local
    variable and when all the files are processed we raise an Exception that
    includes all of them).
    """
    # Execute the read function
    read = self.__readFile(self.filename)
    # Raise an error if the file was not succesfully read
    if not read:
      # Raise the error leaving the object clean
      if self.__dict__.has_key('_errors'):
        msgs = self._errors
        raise ControlFileError(msgs)
    # Return if the config file was succesfully read
    return read
  
  def _parseParagraphs(self):
    """
    Fucntionan that parses the paragraphs read by the _readFile function and
    generates a dictionary of tasks that is stored on the class.
    """
    # Execute the read function
    parsed = self.__parseParagraphs()
    # Raise an error if the file was not succesfully read
    if not parsed:
      # Raise the error leaving the object clean
      if self.__dict__.has_key('_errors'):
        msgs = self._errors
        raise ControlFileError(msgs)
    # Return if the config file was succesfully parsed
    return parsed
    
  def __begRead(self):
    """Private function to initialize global variables used by _ReadFile"""
    self._filesBeingRead = []
    self._seenFiles      = []
    self._fields         = []
    self._inpara         = False
    # Save working directory
    self._workdir        = os.getcwd()

  def __endRead(self):
    """Private function to cleanup global variables used by _ReadFile"""
    # Return to the original working directory, if saved
    if self.__dict__.has_key('_workdir'):
      os.chdir(self._workdir)
    # Remove all local variables created on _begRead
    for k in ('_workdir', '_filesBeingRead', '_seenFiles', '_fields', '_inpara'):
      if self.__dict__.has_key(k):
        del self.__dict__[k] 

  def __logReadError(self, etype, msg):
    """
    Function used to record read errors; if ``stopOnFirstError`` is True the
    function raises the error
    """
    err = (etype, _("Read ERROR: %s") % (msg))
    if self.stopOnFirstError:
      self.__endRead()
      raise ControlFileError([err,])
    else:
      if not self.__dict__.has_key('_errors'):
        self._errors = []
      self._errors.append(err)

  def __readFile(self, filename):
    """
    Function that reads the control file and the files it includes and
    generates a list of paragraphs.

    Each paragraph is a list of the form (field, value, reference), where
    field is the field name, value the contents after the colon and reference
    is a string of the form 'filename:linenum' that can be used when analyzing
    the paragraphs.
    """
    # Use absolute paths for files
    filename = os.path.abspath(filename)
    # Check if we are on the first iteration and don't read already seen files
    # to avoid file inclusion loops
    if not self.__dict__.has_key('_filesBeingRead'):
      self.__begRead()
    elif filename in self._filesBeingRead:
      return True
    elif filename in self._seenFiles:
      return True
    # Try to open the file
    try:
      f = open(filename, 'r')
    except IOError:
      self.__logReadError(
        eType.IOError,
        _("Unable to open file '%s'") % filename
      )
      return False
    # Append file to files being read
    self._filesBeingRead.append(filename)
    # And start processing it
    lineno = 0
    for line in f.xreadlines():
      lineno += 1
      # Remove trailing spaces on the lines
      line = string.rstrip(line)
      # Test empty lines (the original empty or only had whitespace)
      if line == "":
        if self._inpara:
          self.paragraphs.append(self._fields)
          self._fields = []
          self._inpara = False
        continue
      # Test comments
      if line[0] == '#':
        continue
      # Test continuation lines
      if line[0] == ' ' or line[0] == '\t':
        if line != '' and self._inpara:
          self._fields[-1][1] += '\n' + line
          continue
        else:
          self.__logReadError(
            eType.InvalidLine,
            _("'%s:%d':Unexpected data '%s'") % (filename, lineno, line)
          )
      # If we are here the line must be a normal data field line
      fv = string.split(line, ':')
      if len(fv) > 1:
        if not self._inpara:
          self._inpara = True
        # Append the line, we lowercase the field name and do a lstrip for the
        # value (the right one was done on the full line ;) )
        fname  = string.lower(fv[0])
        fvalue = string.lstrip(fv[1])
        self._fields.append([fname, fvalue, filename, lineno])
        if fname in inclKeywords:
          # Save old errors, to avoid mixing them with new ones
          if self.__dict__.has_key('_errors'):
            old_errors = self._errors
            del self.__dict__['_errors']
          else:
            old_errors = None
          # Change working directory to the current file's dir for the new
          # call to _readFile
          oldwd = os.getcwd()
          os.chdir(os.path.dirname(filename)) # filename is an absolute path
          # Read the file
          result = self.__readFile(fvalue)
          # Go back to previous working directory
          os.chdir(oldwd)
          # Test result and recover from old errors
          if not result:
            new_errors   = self._errors
            if old_errors != None:
              self._errors = old_errors
            else:
              del self._errors
            self.__logReadError(
              eType.IncludeError,
              _("'%s:%d':Error(s) reading included file '%s'") % (filename, lineno, fvalue)
            )
            self._errors += new_errors
          elif old_errors != None:
            self._errors = old_errors
      else:
        self.__logReadError(
          eType.InvalidLine,
          _("'%s:%d':Error reading field '%s'") % (filename, lineno, line)
        )
    # Close file
    f.close()
    # Remove it from the list of files beeing read and put it on the seen
    # files list
    self._filesBeingRead.remove(filename)
    self._seenFiles.append(filename)
    # Operations when we have read all files
    if len(self._filesBeingRead) == 0:
      # Append the last paragraph if we are on the last file
      if self._inpara:
        self.paragraphs.append(self._fields)
      # Cleanup the local variables and go back to the original directory
      self.__endRead()
    # Return True if reading was successful
    if self.__dict__.has_key('_errors'):
      return False
    else:
      return True
  
  def __begParse(self):
    """Private function to initialize global variables used by __ParseFile"""
    # Save working directory
    self._workdir        = os.getcwd()
  
  def __endParse(self):
    """Private function to cleanup global variables used by __ParseFile"""
    # Return to the original working directory, if saved
    if self.__dict__.has_key('_workdir'):
      os.chdir(self._workdir)
  
  def __logParseError(self, etype, msg):
    """
    Function used to record parse errors; if ``stopOnFirstError`` is True the
    function raises the error
    """
    err = (etype, _("Parse ERROR: %s") % (msg))
    if self.stopOnFirstError:
      self.__endParse()
      raise ControlFileError([err,])
    else:
      if not self.__dict__.has_key('_errors'):
        self._errors = []
      self._errors.append(err)
  
  def __parseValue(self, value, type, file, lineno):
    """
    Function that parses a value depending on the keyword type
    """
    ref = "%s:%d" % (file, lineno)
    ret = None
    if type == kType.bool:
      lv = string.lower(value)
      if lv in ('1', 'true', 'yes', 'on'):
        ret = True 
      elif lv in ('0', 'false', 'no', 'off'):
        ret = False
      else:
        self.__logParseError(
          eType.ValueError,
          _("'%s':Value '%s' is not boolean") % (ref, value)
        )
    elif type == kType.fpath:
      oldwd = os.getcwd()
      os.chdir(os.path.dirname(file)) # file is an absolute path
      ret = os.path.abspath(value)
      os.chdir(oldwd)
    elif type == kType.num:
      try:
        ret = int(value)
      except ValueError:
        self.__logParseError(
          eType.ValueError,
          _("'%s':Value '%s' must be an integer") % (ref, value)
        )
    elif type == kType.pdep:
      ret = apt_pkg.ParseDepends(value)
    elif type == kType.plist:
      pl = map(string.strip, string.split(value, ','))
      for pn in pl:
        if not chkPackageName(pn):
          self.__logParseError(
            eType.ValueError, _("'%s':Invalid package name '%s'") % (ref, pn)
          )
      ret = pl
    elif type == kType.pname:
      if chkPackageName(value):
        ret = value
      else:
        self.__logParseError(
          eType.ValueError, _("'%s':Invalid package name '%s'") % (ref, value)
        )
    elif type == kType.str:
      ret = value
    elif type == kType.slist:
      sl = map(string.strip, string.split(value, ','))
      for sn in sl:
        if not chkScriptName(sn):
          self.__logParseError(
            eType.ValueError, _("'%s':Invalid script name '%s'") % (ref, pn)
          )
      ret = sl
    elif type == kType.sname:
      if chkScriptName(value):
        ret = value
      else:
        self.__logParseError(
          eType.ValueError, _("'%s':Invalid script name '%s'") % (ref, value)
        )
    else:
      self.__logParseError(
        eType.TypeError,
        _("'%s':Unknown keyword type '%s' for value '%s'") % (type, value)
      )
    return ret
    
  def __parseParagraphs(self):
    """Parse paragraphs and generate a dictionary of tasks"""
    self.__begParse()
    # Local variables
    # Current task
    cTask = None
    # Start processing
    for para in self.paragraphs:
      # A new paragraph starts on cddd scope and the current dictionary is the
      # global one
      cScope = kScope.cddd
      cDict  = self.globals
      # Parse paragraph
      for key, val, file, lineno in para:
        ref = "%s:%d" % (file, lineno)
        # Test if we are reading an unknown keyword
        if not keywords.has_key(key):
          # save x- keywords
          if key[0] == 'x' and key[1] == '-':
            if cDict.has_key(key):
              cDict[key].append(val)
            else:
              cDict[key] = [ val ]
          else:
            self.__logParseError(
              eType.InvalidKeyword,
              _("'%s':Unknown keyword '%s'") % (ref, key)
            )
          continue
            
        # Ignore include keywords ... they are already procesed
        if key in inclKeywords:
          continue
        kd = keywords[key]
        # Test if the key is valid on this scope
        if cScope not in kd['scope']:
          self.__logParseError(
            eType.InvalidKeyword,
            _("'%s':Keyword '%s' not allowed on scope '%s'") %
            (ref, kd['name'], cScope)
          )
          continue
        # Parse the value
        val = self.__parseValue(val, kd['type'], file, lineno)
 
        # Get key scopes
        ksc = kd['scope'][cScope]
        # Do things depending on the key we have
        if key in taskKeywords: 
          if not self.tasks.has_key(val):
            # New task, set value
            self.tasks[val] = CDDTask(key, val)
            self.tasks[val][key] = val
          # By default we append things to the task
          cTask = val
          cDict = self.tasks[cTask]
        else:
          # Standard key
          if cDict.has_key(key):
            if ksc['dup'] == kDup.append:
              cDict[key].append(val)
            elif ksc['dup'] == kDup.concat:
              cDict[key] += val
            elif ksc['dup'] == kDup.fail:
              self.__logParseError( 
                eType.InvalidKeyword,
                _("'%s':Duplicate keyword '%s' not allowed here") %
                (ref, kd['name'])
            )
            elif ksc['dup'] == kDup.replace:
              cDict[key] = val
            # ignore is the default
          else:
            if ksc['dup'] == kDup.append:
              cDict[key] = [ val ]
            else:
              cDict[key] = val
        # Set next scope
        if ksc.has_key('next'):
          cScope = ksc['next']
    # End parsing
    self.__endParse()
    # Return True if parsing was successful
    if self.__dict__.has_key('_errors'):
      return False
    else:
      return True

# ------
# SVN Id: $Id: CtrlFile.py 235 2006-01-29 22:14:54Z sto $
