# debpartial-mirror - partial debian mirror package tool
# (c) 2004 Otavio Salvador <otavio@debian.org>, Nat Budin <natb@brandeis.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston MA 02110-1301 USA
#
# $Id: PackageList.py 310 2006-02-04 10:46:37Z sto $
"""
PackageList classes and functions

This file implements a class to represent lists of packages and operations
related to those lists, like dependencies resolution and package filtering
"""

import re
import os

import apt_pkg

class PackageAlreadyExists(Exception):
    """
    Exception called when someone tries to add a package but it was already
    included.
    
    Parameters:

    - *package*: The name of package.
    """
    def __init__(self, package):
        Exception.__init__(self)
        self.package = package
        
class PackageDoesNotExist(Exception):
    """
    Exception called when someone tries to remove a package that was not
    included.

    Parameters:

    - *package*: The name of package.
    """
    def __init__(self, package):
        Exception.__init__(self)
        self.package = package

class BrokenDependencies(Exception):
    """
    Exception called when, after trying to solve dependencies, we are left with
    broken ones.

    Parameters:

    - *deps*: List of unresolved dependencies.
    """
    def __init__(self, deps):
        Exception.__init__(self)
        self.deps = deps
    
class InvalidFilter(Exception):
    """
    Exception called when someone try to use a invalid filter option.

    Parameters:

    - *option*: The filter option.
    """
    def __init__(self, option):
        Exception.__init__(self)
	self.option = option

class PackageList:
    """
    This class is used to store a list of packages and provide ways to select
    subsets of those packages applying filters to the list.
    
    It also provides methods to resolve depencencies (Depends, Recommends and
    Suggests) against a full set of packages and include the needed packages
    on the list.

    To make filtering faster the class includes caches by name, subsection and
    priority.
    """

    def __init__(self):
        self._name = {}
        self._subsection = {}
        self._priority = {}
        self._provides = {}
    
    def __len__(self):
        return len(self._name)

    def __getitem__(self, key):
        if not self._name.has_key(key):
            raise PackageDoesNotExist, key
        return self._name[key]

    # allow uses like ... if x in y:
    def __contains__(self, item):
        return self._name.has_key(item)

    def has_key(self, key):
        return self._name.has_key(key)

    def keys(self):
        """Returns a copy of the list of package names"""
        return self._name.keys()
        
    def values(self):
        """Returns a copy of the list of packages"""
        return self._name.values()

    def items(self):
        """Returns a copy of the list of (name, package) pairs"""
        return self._name.items()

    def get(self, key):
        """Returns a package given it's name"""
        if not self._name.has_key(key):
            raise PackageDoesNotExist, key
        return self._name[key]
        
    def add(self, package):
        """
        Method to add a package to the list (updating caches).
        
        Note that the package is only added if it's name is not already on the
        list of packages, or else a PackageAlreadyExists exception is raised.
        """
        if self._name.has_key(package['Package']):
            raise PackageAlreadyExists, package['Package']

        if not self._subsection.has_key(package['Section']):
            self._subsection[package['Section']] = []
        if not self._priority.has_key(package['Priority']):
            self._priority[package['Priority']] = []

        self._name[package['Package']] = package
        self._subsection[package['Section']].append(package)
        self._priority[package['Priority']].append(package)

        if package.has_key('Provides'):
            for p in package['Provides']:
                if not self._provides.has_key(p):
                    self._provides[p] = []
                self._provides[p].append(package)

    def remove(self, package):
        """
        Method to remove a package from the list (updating caches).

        If the package was not on the list a PackageDoesNotExist exception is
        raised.
        """
        if isinstance(package, str):
            if not self._name.has_key(package):
                raise PackageDoesNotExist, package
            package = self._name[package]
        else:
            if not self._name.has_key(package['Package']):
                raise PackageDoesNotExist, package['Package']
            
        del self._name[package['Package']]

        self._subsection[package['Section']].remove(package)
        self._priority[package['Priority']].remove(package)
            
        if package.has_key('Provides'):
            for p in package['Provides']:
                if not self._provides.has_key(p):
                    self._provides[p] = []
                self._provides[p].remove(package)

    def resolveDepends(self, pkglist):
        """
        Resolve our dependencies adding needed packages to our list from the
        `pkglist`.
        """
        from PackageListOps import getPkgListDepends
        (depDict, broken) = getPkgListDepends(self.values(), self, [pkglist,])
        for p in depDict.keys():
          self.add(depDict[p][p])
        if broken != None:
          raise BrokenDependencies(broken)

    def resolveRecommends(self, pkglist):
        """
        Add all the recommended packages and the dependencies to our list taking
        missing ones from `pkglist`. 
        """
        from PackageListOps import getPkgListRecommends
        (depDict, broken) = getPkgListRecommends(self.values(), self, [pkglist,])
        for p in depDict.keys():
          self.add(depDict[p][p])
        if broken != None:
          raise BrokenDependencies(broken)
            
    def resolveSuggests(self, pkglist):
        """
        Add all the suggested packages and the dependencies to our list taking
        missing ones from `pkglist`. 
        """
        from PackageListOps import getPkgListSuggests
        (depDict, broken) = getPkgListSuggests(self.values(), self, [pkglist,])
        for p in depDict.keys():
          self.add(depDict[p][p])
        if broken != None:
          raise BrokenDependencies(broken)

    def filter(self, condition):
        """
        Return a PackageList that contains only the packages that verify a set
        of conditions.

        The condition argument is a dictionary that can have the following keys:

        - *include-from*: includes packages form the given debian-cd task
        
        - *exclude-from*: excludes packages form the given debian-cd task
        
        - *subsection*: includes packages that belong to a section that match
          the given regular expression
          
        - *priority*: includes packages that have a priority that match the
          given regular expression

        - *name*: includes packages that have a name that match the given
          regular expression.
          
        If the condition was not on the above list a InvalidFilter exception is
        raised.
        """
        packages = PackageList()
        
        for (item, value) in condition.items():
            # Handle tasks files
            if item in ("include-from", "exclude-from"):
                #print "Processing task (%s): %s" % (item, value)
                # FIXME: Need handle by arch
                raw = os.popen("cpp -I %s -DARCH_i386 %s | grep -v \'^$\|^#.*$\' | tr \'\n\' \' \'"
                               % (os.path.dirname(value), value), 'r' ).readlines()[0].split()

                for p in raw:
                    pkg = self._name.get(p, None)
                    if item == 'include-from':
                        if pkg and pkg['Package'] not in packages: packages.add(pkg)
                    else:
                        if pkg and pkg['Package'] in packages: packages.remove(pkg)
                
                # Process new item
                continue
            
            # Handle catalog
            if item == "subsection":
                d = self._subsection
            elif item == "priority":
                d = self._priority
            elif item == "name":
                d = self._name
            else:
                raise InvalidFilter(item)

            regexp = re.compile(value)
            for k in d.keys():
                if k and regexp.match(k):
                    # Add all match packages
                    if isinstance(d[k], list):
                        for i in d[k]:
                            try:
                                packages.add(i)
                            except PackageAlreadyExists:
                                pass # We possible will repeat packages
                    else:
                        try:
                            packages.add(d[k])
                        except PackageAlreadyExists:
                            pass # We possible will repeat packages

        return packages
