"""
Package List Operations.

This functions operate over PackageLists.
"""

# TODO: Add functions to support CDDTasks

import apt_pkg
import sys

DEBUG = True

# Helper Functions
def __getPkgDeps(pkg, depFields):
  """
  Return the list of dependencies of a package
  """
  pDeps = []
  for field in depFields:
    if pkg.has_key(field):
      pDeps += pkg[field]
  return pDeps 
  
def __virtPkgToDeps(origDeps, fullPkgListSet):
  """
  Review all packages on the origDep and replace the virtual packages by an
  or'ed list of real packages
  """
  newDeps = []
  for dep in origDeps:
    ndep  = []
    for (pkgname, pkgver, cond) in dep:
      found = False
      for pl in fullPkgListSet:
        if pkgname in pl._provides:
          # First, if there is a non-vitual package, add it
          if pkgname in pl:
            ndep.append( (pkgname, pkgver, cond) )
          # And later add the packages that provide the package
          for ppkg in pl._provides[pkgname]:
            ndep.append( (ppkg['Package'], '', '') )
          found = True
      # It wasn't a virtual package
      if not found:
        ndep.append( (pkgname, pkgver, cond) )
    newDeps.append(ndep)
  return newDeps

def __chkPkgListDeps(pkgList, currentPkgList, fullPkgListSet,
                     depFields=('Depends','Pre-Depends')):
  """
  Return list of dependencies of the packages on the pkgList that are not
  found on the currentPkgList
  """
  pendingDeps = []
  for pkg in pkgList:
    # Put on pkgDeps the list of dependencies (each dependency is a lists of tuples)
    pkgDeps = __getPkgDeps(pkg, depFields)
    # Each dependency is a list of tuples where only one of them has to be satisfied
    # (it is a list of alternate dependencies)
    for dependency in pkgDeps:
      resolved = False
      # We check each dependency agaist the original package list
      for (pkgname, pkgver, cond) in dependency:
        if pkgname in currentPkgList:
          package_version = pkg['Version']
          if apt_pkg.CheckDep(package_version, cond, pkgver):
            resolved = True
            break
      # If the dependency is not verified using the currentPkgList we put it on
      # of the lists of unmet dependencies (simple ones and alternatives)
      if not resolved and dependency not in pendingDeps:
        pendingDeps.append(dependency)
  return __virtPkgToDeps(pendingDeps, fullPkgListSet)

def __getPkgListDeps(pkgList, currPkgList, fullPkgListSet, 
                     depFields=('Depends','Pre-Depends')):
  """
  This function works as __getDepsListDeps, but the dependency list is built
  using the **pkgList** parameter, which lists the packages we want to resolve
  depependencies for.
  """
  depList = __chkPkgListDeps(pkgList, currPkgList, fullPkgListSet, 
                              depFields)
  return __getDepListDeps(depList, currPkgList, fullPkgListSet, depFields)
   
def __getDepListDeps(depList, currPkgList, fullPkgListSet, 
                     depFields=('Depends','Pre-Depends')):
  """
  Parameters:
  
  - **depList**: List of dependencies to satisfy
  
  - **currentPkgList**: The PackageList we want to complete to be able to
    install the pkgList.
  
  - **fullPkgListSet**: A list of PackageList (usually full lists). The order
    of the list is the order that is used to find the packages.
  
  Returns:

  - A tuple (dict, list) where the dictionary is a mapping between the names
    of the needed packages and the PackageList they come from and the list is
    None if there are no unresolved dependencies or a list of tuples of the
    form (pkgname, pkgver, cond).

  Given a list of dependencies, a list of already available packages and a
  list of PackageLists return the minimum set of packages needed to satisfy
  the dependencies.

  After the original fields are tested we use 'Depends' and 'Pre-Depends' only.

  When all the dependencies are satisfied the function returns a tuple:
  
  - the first element is a dictionary where the keys are the packages we need
    to get and the values are a reference to the PackageList it comes from. 
  
  - the second element is a list of unsatisfied dependencies (pkgname, pkgver,
    cond), if there are any.
  """
  # Dictionary of new packages
  newPkgs = {}
  # or'ed dependencies
  altDepsPending = []
  # resolved dependencies
  resolvedDeps = []
  # unresolved dependencies
  unresolvedDeps = []
  # First, build a list of dependecies not included on the currentPkgList
  pendingDeps = depList
  # Now, we will calculate the minimum set of packages needed to solve the
  # dependencies, we will do it in two loops:
  #
  # 1. First we will evaluate simple dependencies recursively, we know that
  #    all simple dependencies have to be satisfied to be able to install, 
  #    on this run all the virtual package dependencies are transformed into
  #    normal dependencies
  #
  # 2. On the second loop we will solve the or'ed dependencies, all the
  #    alternatives will be evaluated against already selected packages
  #    and if none matches we will try to add a package that solves the
  #    first dependency

  # First loop: simple dependencies
  while len(pendingDeps) > 0:
    addedPackages  = []
    for dep in pendingDeps:
      # We are only evaluating simple dependencies, move or'ed ones
      if len(dep) != 1:
        altDepsPending.append(dep)
        continue
      # See if the dependency has already been solved
      if dep in resolvedDeps:
        continue
      # Get simple dependency values
      (pkgname, pkgver, cond) = dep[0]
      # First see if the package is on the newPkgs list (in that case the
      # package and it's dependencies are already being considered)
      if pkgname in newPkgs:
        package_version = newPkgs[pkgname][pkgname]['Version']
        if apt_pkg.CheckDep(package_version, cond, pkgver):
          resolvedDeps.append(dep)
          continue
      # Now look on the full PackageListSet, check if there is a package that
      # resolves the dependency. If the package exists add it to newPkgs list
      # and to this iteration's list
      resolved = False
      for pl in fullPkgListSet:
        if pkgname in pl:
          package_version = pl[pkgname]['Version']
          if apt_pkg.CheckDep(package_version, cond, pkgver):
            resolved         = True
            newPkgs[pkgname] = pl
            addedPackages.append(pl[pkgname])
            resolvedDeps.append(dep)
            break
      # If the dependency was not solved, the needed package does not exist
      if not resolved and dep not in unresolvedDeps:
        unresolvedDeps.append(dep)
    # Evaluate the new package depends on the next iteration
    pendingDeps = __chkPkgListDeps(addedPackages, currentPkgList, fullPkgListSet)

  # Second loop: check multiline deps
  pendingDeps  = altDepsPending
  while len(pendingDeps) > 0:
    for dep in pendingDeps:
      if dep in resolvedDeps:
        continue
      for (pkgname, pkgver, cond) in dep:
        if pkgname in newPkgs:
          package_version = newPkgs[pkgname][pkgname]['Version']
          if apt_pkg.CheckDep(package_version, cond, pkgver):
            resolvedDeps.append(dep)
            break
    # Now, check against the fullPkgListSet
    addedPackages = []
    for dep in pendingDeps:
      if dep in resolvedDeps:
        continue
      resolved = False
      for (pkgname, pkgver, cond) in dep:
        for pl in fullPkgListSet:
          if pkgname in pl:
            package_version = pl[pkgname]['Version']
            if apt_pkg.CheckDep(package_version, cond, pkgver):
              resolved         = True
              newPkgs[pkgname] = pl
              addedPackages.append(pl[pkgname])
              break
        if resolved:
          break
      # If the dependency was not solved, the dependency can't be satisfied
      if not resolved and dep not in unresolvedDeps:
        unresolvedDeps.append(dep)
    # Evaluate the new package depends on the next iteration
    pendingDeps = __chkPkgListDeps(addedPackages, currentPkgList, fullPkgListSet)
  # Return None if there are no unresolved dependencies
  if len(unresolvedDeps) == 0:
    unresolvedDeps = None
  return (newPkgs, unresolvedDeps)

####################
# Public functions #
####################

def getPkgListDepends(pkgList, currentPkgList, fullPkgListSet):
  return __getPkgListDeps(pkgList, currentPkgList, fullPkgListSet,
                          ('Depends', 'Pre-Depends'))

def getPkgListRecommends(pkgList, currentPkgList, fullPkgListSet):
  return __getPkgListDeps(pkgList, currentPkgList, fullPkgListSet,
                          ('Depends', 'Pre-Depends', 'Recommends'))

def getPkgListSuggests(pkgList, currentPkgList, fullPkgListSet):
  return __getPkgListDeps(pkgList, currentPkgList, fullPkgListSet,
                          ('Depends', 'Pre-Depends', 'Recommends', 'Suggests'))

def getPkgListBuildDepends(srcList, currentPkgList, fullPkgListSet):
  return __getPkgListDeps(srcList, currentPkgList, fullPkgListSet,
                          ('Build-Depends', 'Build-Depends-Indep'))

def srcPkgListFromBinPkgList(sourcesPkgList, binPkgList):
  srcList = []
  binMap = {}
  for s in sourcesPkgList.values():
    for b in s['Binary']:
      binMap[b]=s
  for p in binPkgList:
    try:
      s = binMap[p]
      if s not in srcList:
        srcList.append(s)
    except PackageList.PackageDoesNotExist:
      pass
  return srcList
