# ---------------------------------------------------
# PyCDDT - Python Custom Debian Distributions Toolkit
# ---------------------------------------------------
# Module:    PyCDDT
# File:      Utils.py
# Author:    Sergio Talens-Oliag <sto@debian.org>
# Copyright: (c) 2005, 2006 Sergio Talens-Oliag <sto@debian.org>
# SVN Id:    $Id: Utils.py 310 2006-02-04 10:46:37Z sto $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston MA 02110-1301 USA
# ---------------------
# -*- coding: utf-8 -*-
# ---------------------

"""
Custom Debian Distributions Toolkit Utilities

This file includes misc classes and functions used on the rest of the files.
"""

import locale
import os.path
import string
import sys

# Enum class: it is a list of
class Enum:
  """Create an enumerated type.
  The constructor takes a list of names (a list or a string with names
  separated by spaces) and creates the objects.
  """
  class Iterator:
    def __init__(self, enum, index=0):
      self.enum  = enum
      self.index = index
    def __iter__(self):
      return self
    def next(self):
      if (len(self.enum) <= self.index):
        raise StopIteration()
      value = self.enum[self.index]
      self.index += 1
      return value
  def __init__(self, names=[]):
    """Create attributes on the object with the given names"""
    if type(names) == type(""):
      self.__names = names.split()
    else:
      self.__names = names
    for var in self.__names:
      if type(var) != type(""):
        raise TypeError( _("Values must be strings") )
      if var in vars(self).keys():
        raise AttributeError( _("Duplicate value in Enum") )
      vars(self)[var]=var
  def __contains__(self, value):
    return value in self.__names
  def __iter__(self):
    return Enum.Iterator(self)
  def __getitem__(self, index):
    return self.__names[index]
  def __len__(self):
    return len(self.__names)
  def __str__(self):
    return str(self.__names)

# Check functions
def chkPackageName(name):
  """
  Return True iff the name is a string that consist only of lower case
  letters (a-z), digits (0-9), plus (+) and minus (-) signs, and periods
  (.). They must be at least two characters long and must start with an
  alphanumeric character.
  """
  # Check length
  if len(name)<2:
    return False
  # Check first character
  if name[0] not in string.ascii_lowercase:
    return False
  # Validate the rest of the name
  valid_characters = string.ascii_lowercase + string.digits + '+-.'
  for c in name[1:]:
    if c not in valid_characters:
      return False
  return True

def chkScriptName(name):
  """
  Return True iff the name is a string that consist entirely of upper and
  lower case letters, digits, underscores, and hyphens.
  """
  valid_characters = string.ascii_letters + string.digits + '_-'
  for c in name:
    if c not in valid_characters:
      return False
  return True

# Auxiliary function to generate relative paths
def mkRelPATH(bpath, dpath):
  """Return the route to dest_path from base_path iff we don't arrive to the
  root of the system; in that later case dpath is returned unchanged.
  """
  if bpath != None:
    bpath = os.path.abspath(bpath) + os.path.sep
    base_path = string.split(bpath, os.path.sep)
    dest_path = string.split(dpath, os.path.sep)
    for i in range(min(len(base_path), len(dest_path))):
      if base_path[i] != dest_path[i]:
        if i>1 and i < len(dest_path):
          new_path = [os.path.pardir] * (len(base_path)-(i+1)) + dest_path[i:]
          return string.join( new_path, os.path.sep )
        break
  return dpath
  
# String functions
def genUnderline(str, chr='-'):
  """
  Generate one line of the same size as the string
  """
  ul = ""
  for i in xrange(len(str)):
    ul += chr
  return ul

def pdepToStr(dep):
  """
  Transform a package dependency into a string
  """
  dlist = []
  for d in dep:
    if d[1] == "" and d[2] == "":
      dlist.append(d[0])
    else:
      dlist.append("%s (%s %s)" % (d[0],d[2],d[1]))
  return string.join(dlist, " | ")

def lpout(s):
  """Print on stdout using the preferred locale encoding"""
  try:
    encoding = locale.getpreferredencoding()
  except locale.Error:
    encoding = 'utf-8'
  print s.encode(encoding)

def lperr(s):
  """Print on stderr using the preferred locale encoding"""
  try:
    encoding = locale.getpreferredencoding()
  except locale.Error:
    encoding = 'utf-8'
  print >> sys.stderr, s.encode(encoding)

# ------
# SVN Id: $Id: Utils.py 310 2006-02-04 10:46:37Z sto $
